//   $id: $
//    ___
//   |  _\ A free Lemmings clone
//   |   /_  _ _  ___  _   _  ___ 
//   |  || || \ || _ || |_| ||_ -'
//   |__||_||_\_||_  ||_____||___|
//                _| |
//               |___|
//
//   Copyright (C) 1998 Ingo Ruhnke <grumbel@gmx.de>
//     
//   This program is free software; you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation; either version 2 of the License, or
//   (at your option) any later version.
//
//   This program is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with this program; if not, write to the Free Software
//   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include <config.h>
#include <iostream>
#include <strstream>
#include <ctime>
#include <cstdio>
#include <unistd.h>
#include <getopt.h>

#include <ClanLib/clanlib.h>
#include <ClanLib/clanlayer2.h>

#include "Client.hh"
#include "AbstractServer.hh"
#include "TrueServer.hh"
#include "Playfield.hh"
#include "ButtonPanel.hh"
#include "Pingus.hh"
#include "Pingu.hh"
#include "algo.hh"
#include "globals.hh"
#include "LevelMap.hh"
#include "actions/pingu_actions.hh"
#include "LevelInterrupt.hh"
#include "SoundSpecs.hh"
#include "GameObj.hh"
#include "PingusError.hh"

using namespace std;

PingusMain::PingusMain()
{
  // default values
  resolution = "640x480";
  title = string(PACKAGE) + " " + VERSION + " - A free Lemmings clone";
  pingus_datadir_set = false;
}

char* 
PingusMain::get_title()
{
  title += " - http://pingus.home.pages.de";
  char* title_str = strdup(title.c_str());
  return title_str;
}

// check_ars() checks the command line for options and set the
// corresponding global variables to the set values.
void 
PingusMain::check_args(int argc, char* argv[])
{
  int c; 
  int option_index = 0;
  //  int display_help = 0;

  struct option long_options[] =
  {
    {"enable-music",    no_argument,       0, 'm'},
    {"enable-sound",    no_argument,       0, 's'},
    {"enable-gimmicks", no_argument,       0, 'i'},
    {"set-speed",       required_argument, 0, 't'},
    {"datadir",         required_argument, 0, 'p'},
    {"level",           required_argument, 0, 'l'},
    {"help",            no_argument,       0, 'h'},
    {"version",         no_argument,       0, 'V'},
    {"verbose",         no_argument,       0, 'v'},
    {"print-fps",       no_argument,       0, 'b'},
    {"sound-specs",     required_argument, 0, 'S'},
    {"geometry",        required_argument, 0, 'g'},
    {0, 0, 0, 0}
  };

  while(true) {
    c = getopt_long(argc, argv, "smvdl:hVf:p:bxS:g:it:", long_options, &option_index);
    
    if (c == -1)
      break;

    switch(c) {
    case 0:
      cout << "You should normaly never reach this line." << endl;
      exit(EXIT_FAILURE);
      break;
    case 'b': // -b, --print-fps
      print_fps = true;
      if (verbose) cout << "check_args: Printing fps enabled" << endl;
      break;
    case 'l': // -l, --level
      levelfile = optarg;
      if (verbose) cout << "check_args: Levelfile = " << levelfile << endl;
      break;
    case 't': // -t, --set-speed
      game_speed = atoi(optarg);
      if (verbose) 
	cout << "check_args: Game Speed = " << game_speed << endl;
      break;
    case 's': // -s, --enable-sound
      if (verbose) cout << "check_args: Sound Effects enabled" << endl;
      sound_enabled = true;
      break;
    case 'g':
      resolution = optarg;
      break;
    case 'S':
      pingus_soundfile = optarg;
      if (verbose) 
	cout << "check_args: Sound File = " << pingus_soundfile << endl;
      break;
    case 'm': // -m, --enable-music
      if (verbose) cout << "check_args: Music enabled" << endl;
      music_enabled = true;
      break;
    case 'p': // -p, --datadir
      pingus_datadir = optarg;
      pingus_datadir_set = true;
      if (verbose)
	cout << "check_args: Pingus Data Dir = " << pingus_datadir << endl;
      break;
    case 'V':
      cout << PACKAGE << " version " << VERSION << endl;
      cout <<
	"\n"
	"Copyright (C) 1998 Ingo Ruhnke <grumbel@gmx.de>\n"
	"There is NO warranty.  You may redistribute this software\n"
	"under the terms of the GNU General Public License.\n"
	"For more information about these matters, see the files named COPYING."
	   << endl;

      exit(EXIT_SUCCESS);
      break;
    case 'i':
      gimmicks_enabled = true;
      if (verbose) cout << "Pingus: Gimmicks enabled" << endl;
      break;
    case 'v':
      ++verbose;
      cout << "Being more verbose: " << verbose << endl;
      break;
    default:
      cout << "Usage: " << argv[0] << " [OPTIONS]..." << endl;
      cout <<
	"\n"
	"Options:\n"
	"   -g, --geometry {width}x{height}\n"
	"                           Set the resolution for pingus (default: 640x480)\n"
	"   -h, --help              Displays this screen\n"
	"   -s, --enable-sound      Enable sound\n"
	"   -m, --enable-music      Enable music\n"
	"   -i, --enable-gimmicks   Enable some buggy development stuff\n"
	"   -S, --sound-specs FILE  Use files mentioned in FILE\n"
       	"   -p, --datadir PATH      Set the path to load the data files to `path'\n"
	"   -l, --level FILE        Load a custom level from `file'\n"
	"   -t, --speed SPEED       Set the game speed (0=fastest, >0=slower)\n"
	"   -b, --print-fps         Prints the fps to stdout\n"
	"   -v, --verbose           Print some more messages to stdout, can be set\n"
	"                           multible times to increase verbosity\n"
	"   -V, --version           Prints version number and exit\n"
	   << endl;
      exit(EXIT_SUCCESS);
      break;
    }
  }

  // Treating non option arguments
  for(int i=optind; i < argc; ++i) {
    cout << "Ignoring, unknown argument: " << argv[i] << endl;
  }
}

void
PingusMain::intro(void)
{
  // Show a simple startup intro
}

// Loads a level from file and starts playing i
Result
PingusMain::play_level(string filename)
{
  Result r;
  Client* client;
  AbstractServer* server;
  PLF* plf;
  
  plf = new PLF(filename);

  server = new TrueServer(plf, r);
  client = new Client(server);

  client->start(filename);

  return r;
}

void
PingusMain::load_data()
{
  if (verbose) {
    cout << "Loading data... " << endl;
    cout << pingus_scriptfile <<  endl;
    cout << pingus_datafile << endl;
  }

  resource = new CL_ResourceManager("pingus.scr",
				    pingus_datafile.c_str());

  GameObj::SetResourceManager(resource);

  Pingu::load_data();
  PinguAction::load_data();
  Particle::load_data();
  World::load_data();
  Playfield::load_data();
  ButtonPanel::load_data();
  LevelMap::load_data();
  Entrance::load_data();
  Trap::load_data();
  Exit::load_data();
  CaptureRectangle::load_data();

  if (verbose)
    cout << "Loading data successfull" << endl;
}

// Get all filenames and directories
void
PingusMain::get_filenames()
{
  if (verbose)
    clog << "Getting filenames..." << endl;

  char* env;
  string hotspot_file;

  // Search different places for the datadir path, if none can be
  // found, then fail.
  if (!pingus_datadir_set) {
    env = getenv("PINGUS_DATADIR");
    
    if (env) {
      if (verbose > 1) cout << "Using envar $PINGUS_DATADIR" << endl;
      pingus_datadir = env;
      pingus_datadir_set = true;
    } 

    if (!pingus_datadir_set) {
      if (verbose > 1) cout << "Using intern macro PINGUS_DATADIR" << endl;
      pingus_datadir = PINGUS_DATADIR;
    }

    // Correct the path name, if no slash is pressent
    add_slash(pingus_datadir);
  
    if (!pingus_datadir_set && exist("../data/pingus.dat")) {
      if (verbose > 1)
	cout << "Assuming that you haven't installed pingus, overriding current value." << endl;
      pingus_datadir = "../data/";
    } 
  }

  if (verbose)
    cout << "pingus_datadir: " << pingus_datadir << endl;

  pingus_datafile   = pingus_datadir + "pingus.dat";
  pingus_scriptfile = pingus_datadir + "pingus.scr";
  hotspot_file      = pingus_datadir + "hotspots.specs";
  
  if (exist(pingus_datafile)) {
    if (verbose > 1)
      cout << "Pingus Datadir exist, all looks ok" << endl;
  } else {
    cout << "Couldn't find `pingus.dat', please set the enviroment variable\n"
	 << "PINGUS_DATADIR to the path of the file `pingus.dat' or use the\n"
	 << "-p option." << endl;
    exit(EXIT_SUCCESS);
  } 

  if (pingus_soundfile == "")
    pingus_soundfile  = pingus_datadir + "sounds/" + "sound.specs";

  if (sound_enabled) {
    cout << __FUNCTION__ << "PingusMain: Reading Sound Specs" << endl;
    SoundSpecs::init(pingus_soundfile);
    cout << "PingusMain: Reading Sound Specs finished" << endl;
  }

  Hotspot::read_info(hotspot_file.c_str());
  
  string custom_levelfile = levelfile;

  if (custom_levelfile == "")
    custom_levelfile = pingus_datadir + "levels/level2.plf";
    
  if (!exist(custom_levelfile)) {
    if (verbose) cout << "Levelfile not found, trying fallbacks" << endl;
    // Search for the level in the datadir
    custom_levelfile = pingus_datadir + "levels/" + custom_levelfile;
    if (!exist(custom_levelfile.c_str())) {
      cerr << "Couldn't find level file: \"" << custom_levelfile << "\"" << endl;
      exit(EXIT_FAILURE);
    } 
  }

  levelfile = custom_levelfile;
  
  if (verbose)
    cout << "Pingus Level File: " << levelfile << endl;
}
  
void
PingusMain::init(int argc, char* argv[])
{
  char c;

  PingusMain::check_args(argc, argv);
  PingusMain::get_filenames();

  // Translate the geometry string to some int's
  if (sscanf(resolution.c_str(), "%d%c%d", &screen_width, &c, &screen_height) != 3) {
    cerr << "Resolution string is wrong, it should be like: \n" 
	 << "\"640x480\" or \"800x600\"" << endl;
    exit(EXIT_FAILURE);
  }

  // initialisation ClanLib
  CL_System::init_display();

  if (sound_enabled || music_enabled)
    CL_System::init_sound();

  if (verbose) {
    cout << "Using resolution: " 
	 << screen_width << "x" << screen_height << endl;
  }

  // Setting the display resolution
  CL_Display::set_videomode(screen_width, screen_height, 16);

  // Loading data and initialisising 
  register_actions();
  load_data();
}

int
PingusMain::main(int argc, char** argv)
{
  Result result;
 
  if (verbose)
    cout << "Starting Main: " << CL_System::get_time() << endl;
  
  try {
    init(argc, argv);  
    intro();
    
    if (verbose) cout << "Levelfile: " << levelfile << endl;

    //    CL_Display::set_videomode(screen_width, screen_height, 16);
    result = play_level(levelfile);
    
    cout << "Pingus exit normaly" << endl;
    result.print();
  }
  
  catch (CL_Error err) {
    cout << "Error caught from ClanLib: " << err.message << endl;
  }

  catch (PingusError err) {
    cout << "Error caught from Pingus: " << err.message << endl;
  }

  return 0;
}

/* EOF */
