// Pingus - A free Lemmings clone
// Copyright (C) 1999 Ingo Ruhnke <grumbel@gmx.de>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

#include <vector>
#include <cstdio>
#include <ctime>

#include "PinguTileMap.hh"
#include "PinguBmpMap.hh"
#include "PinguSpotMap.hh"
#include "PinguRandomMap.hh"

#include "algo.hh"
#include "SoundSpecs.hh"
#include "globals.hh"
#include "Background.hh"
#include "World.hh"
#include "traps/traps.hh"
#include "Result.hh"
#include "actions/pingu_actions.hh"

//CL_ResourceManager* World::resource;

World::World()
{
  
  //event();
  released_pingus = 0;
  colmap = new ColMap;
}

World::~World()
{
  // Bug: This crashes, if pingus is exited
  // delete map;
  // delete  bg;
}

void 
World::draw_pingus(int x_offset, int y_offset, double s=1.0) const
{
  for(unsigned int i=0; i < pingu.size(); ++i) {
    if (pingu[i]->get_status() != dead)
      pingu[i]->draw_offset(x_offset, y_offset, s);
  }
}

// draw the basic background pattern
void 
World::draw_background(int x_offset, int y_offset, double s=1.0) const
{
  bg->draw_offset(x_offset, y_offset, s);

  for(unsigned int i=0; i < hotspot_bg.size(); ++i) 
    hotspot_bg[i]->draw_offset(x_offset, y_offset, s);
			       
}

void 
World::draw_playground(int x_offset, int y_offset, double s=1.0) const
{
  // Draw the gfx map
  if (map)
    map->draw_offset(x_offset, y_offset, s);

  // Draw the exits
  for(unsigned int i=0; i < exits.size(); ++i) {
    exits[i]->draw_offset(x_offset, y_offset, s);
  }
  
  // Draw the entrances
  for(unsigned int i=0; i < entrance.size(); ++i) {
    entrance[i]->draw_offset(x_offset, y_offset, s);
  }

  // Draw the hotspots
}


void 
World::draw_foreground(int x_offset, int y_offset, double s=1.0) const
{
  for(unsigned int i=0; i < traps.size(); ++i) 
    traps[i]->draw_offset(x_offset, y_offset, s);

  for(unsigned int i=0; i < hotspot_fg.size(); ++i) {
    hotspot_fg[i]->draw_offset(x_offset, y_offset, s);
  }
  
  particle.draw_offset(x_offset, y_offset, s);
}

// Merge the different layers on the screen together
void 
World::draw_offset(int x, int y, double s = 1.0) const
{
  draw_background(x, y, s);
  draw_playground(x, y, s);
  draw_pingus(x, y, s);
  draw_foreground(x, y, s);
}

void 
World::let_move()
{
  // Create new pingus, if enough time is passed
  for(unsigned int i=0; i < entrance.size(); ++i) {
    if (pingu.size() <= allowed_pingus
	&& entrance[i]->pingu_ready()) 
      pingu.push_back(entrance[i]->get_pingu());
    
  }
  
  // Let all pingus move
  for(unsigned int i=0; i < pingu.size(); ++i) {
    pingu[i]->let_move();
  }

  // Let the pingus catch each other
  for(unsigned int i=0; i < pingu.size(); ++i) {
     if (pingu[i]->need_catch()) {
    	for(unsigned int j=0; j < pingu.size(); ++j){
		pingu[i]->catch_pingu(pingu[j]);
      	}
     }
  }

  // Let the traps catch the pingus
  for(unsigned int i=0; i < pingu.size(); ++i) {
    for(unsigned int j=0; j < traps.size(); ++j) {
      traps[j]->catch_pingu(pingu[i]);
    }
  }

  for(unsigned int i=0; i < traps.size(); ++i) {
    traps[i]->let_move();
  }

  // Let the exit catch the pingus
  for(unsigned int j=0; j < exits.size(); ++j) {
    for(unsigned int i=0; i < pingu.size(); ++i) {
      exits[j]->catch_pingu(pingu[i]);
    }
  }

  particle.let_move();
}

void 
World::save()
{
  // Empty should save the state of the wolrd 
}

void 
World::load()
{
  // Empty should load the state of the world
}

void
World::load_data()
{
  //  resource = res;
}

void 
World::init(PLF* level_data, Result& r)
{
  result = &r;

  if (verbose)
    cout << "Wolrd init" << endl;

  colmap->load(level_data->get_mapfile());
  
  if (verbose)
    cout << "Loaded Colmap" << endl;

  // load the foreground map
  switch (level_data->map_type()) {
  case TILE:
    map = new PinguTileMap(level_data, resource());
    break;
  case BMP:
    map = new PinguBmpMap(level_data->get_fg(), resource());
    break;
  case SPOT:
    map = new PinguSpotMap(level_data->get_fg(), resource());
    break;
  case RANDOM:
    map = new PinguRandomMap;
    break;
  default:
    cout << "Undef Maptype" << endl;
    exit(EXIT_FAILURE);
  }
  {
    ColMap* temp = map->get_colmap();
    if (temp) {
      delete colmap;
      colmap = temp;
      cout << "Using AutoColMap" << endl;
    } else {
      cout << "Using StaticColMap" << endl;
    }
  }
  
  if (verbose)
    cout << "Loaded pingu map" << endl;

  // load the background map
  bg = new Background(level_data->get_bg());

  if (verbose)
    cout << "Loaded Background" << endl;

  allowed_pingus = level_data->get_pingus();
  result->set_total(allowed_pingus);

  //  int pal;
  //  srand(time(NULL));

  // Creating Exit and Entrance
  vector<exit_data>     exit_d     = level_data->get_exit();
  vector<entrance_data> entrance_d = level_data->get_entrance();

  for(unsigned int i=0; i < exit_d.size(); ++i) 
    exits.push_back(new Exit(exit_d[i]));
  
  for(unsigned int i=0; i < entrance_d.size(); ++i) 
    entrance.push_back(new Entrance(entrance_d[i]));

  vector<trap_data>     trap_d     = level_data->get_traps();
  
  for(unsigned int i=0; i < trap_d.size(); ++i)
    traps.push_back(get_trap(trap_d[i]));

  // Creating the foreground and background hotspots
  vector<hotspot_t> hspot_fg = level_data->get_hotspot_fg();
  vector<hotspot_t> hspot_bg = level_data->get_hotspot_bg();

  for(unsigned int i=0; i < hspot_bg.size(); ++i)
    hotspot_bg.push_back(new Hotspot(hspot_bg[i], resource()));

  for(unsigned int i=0; i < hspot_fg.size(); ++i)
    hotspot_fg.push_back(new Hotspot(hspot_fg[i], resource()));
 
  // Setting Pingus at random positions
  //  pingu.resize(0);

  //  for(int i=0; i < pingu.size(); ++i) 
  //    pingu[i] = new Pingu(rand() % 800, rand() % 300);

  Pingu::set_colmap(colmap);
  Pingu::set_map(map);
  PinguData::SetParticleHolder(&particle);
  Trap::SetParticleHolder(&particle);

  if (verbose)
    cout << "Finished Word init" << endl;

  if (sound_enabled) SoundSpecs::play(SoundSpecs::LETSGO);
}

vector<Pingu* >*
World::get_pingu_p()
{
  return &pingu;
}

int
World::get_width()
{
  return map->get_width();  
}

int
World::get_height()
{
  return map->get_height();
}

void
World::write_results()
{
  result->set_total(allowed_pingus);
  result->set_saved(5);
  result->set_killed(10);
}

void 
World::armageddon(void)
{
  for(unsigned int i=0; i < pingu.size(); ++i)
    pingu[i]->set_action(new Bomber, rand() % 25 + 25);
}

/* EOF */
